/**
 * Copyright 2005 Simon Lehmayr
 *
 * This file is part of Asteroid Zone. Asteroid Zone is free software;
 * you can redistribute it and/or modify it under the terms of the
 * GNU General Public License as published by the Free Software Foundation;
 * either version 2 of the License, or (at your option) any later version.
 * Asteroid Zone is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * You should have received a copy of the GNU General Public License
 * along with Asteroid Zone; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 */

package asteroids;

import javax.microedition.midlet.*;
import javax.microedition.lcdui.*;
import java.util.*;

/**
 * The MIDlet's main class. Handles the game lifecycle and owns
 * the various game screens and commands.
 * @author Jean-Francois Doue
 * @author Simon Lehmayr
 * @version 1.6, 2005/12/20
 */
public class Game extends MIDlet implements CommandListener {
    /** The random number generator used throughout the game */
    public static Random random = new Random();

    /** A timer used to schedule the transitions between the various game screens. */
    public static Timer timer;

    /** The game's main screen, where all the drawing occurs. */
    public static Field field;

    /** The game's sound system */
    public static Soundsystem sound;

    /** A form used to get the player's initials for high scores. */
    public static Form scoreForm;

    /** A form used to display help info. */
    public static Form helpForm;
    
    /** The high score database. */
    public static Scores scores;

    /** The MIDlet display. */
    public static Display display; 
    private static Command _exitCommand;
    private static Command _startCommand;
    private static Command _muteCommand;
    private static Command _musicMuteCommand;
    private static Command _helpCommand;
    private static Command _difficultyCommand;
    private static Command _scoreOkCommand;
    private static Command _helpOkCommand;
    private static TextField _scoreField;
    private static Displayable _currentDisplayable;

    /**
     * Constructor invoked by the application management software.
     * Do not obfuscate.
     */
    public Game() {

        // Retrieve the display.
        display =  Display.getDisplay(this);

        // Create a form which will be used to display
        // licensing info.
        helpForm = new Form("Help");
        helpForm.append("Asteroid Zone - the Return of the Asteroid Blaster\n");
        helpForm.append("This game takes you back to the time when you were addicted to the great arcade game!\nNow available on your java-enabled mobile phone!\n\n");
        helpForm.append("The SETTINGS:\n");
        helpForm.append("Sound is on/off - Enjoy the explosions and mighty engine sounds!\nMusic is on/off - Enjoy the background music\nDifficulty = easy/normal/hard - Change the game according to your personal skill\n\n");
        helpForm.append("The DIFFICULTY SETTINGS:\n");
        helpForm.append("EASY:\nAsteroids have only two sizes. Big ones become two smaller ones when hit. Smaller asteroids explode when hit.\nYour ship has more shields and a faster cannon with a greater shooting range.\nThe asteroids are a bit slower.\nThe ship is invincible for a quite long time after appearing in the asteroid field.\nHalf points for hitting objects in space.\nYou'll get an extra ship every 500 points.\n\nNORMAL:\nAsteroids have three sizes. Every hit will turn a big or medium one into two smaller ones.\nThe UFO appears in level 2 or higher and tries to shoot you with its homing missile.\nYour cannon has normal fire rate and range.\nYour shields are normal.\nNormal points when hitting something.\nYou'll get an extra ship every 1000 points.\n\nHARD:\nLess shields, shooting speed and range.\nThe UFO appears more quickly, tries to shoot you more often and is moving faster.\nThe homing missile is faster.\nThe invincibility time is quite short!\n1.5 times the points when hitting something.\nYou'll get an extra ship every 1500 points.\n\n");
        helpForm.append("The SHIP CONTROLS:\n");
        helpForm.append("You can rotate, accelerate, decelerate and fire using your mobile phone's joystick/cursor navigation.\nActivate the shields with the A game button.\nDo a teleport with the B game button.\n(To locate the game buttons, read your phone's manual!)\n\n");
        helpForm.append("The SHIELDS:\n");
        helpForm.append("You can protect the ship using your shields. Everytime you press the shield button the shields go up, unless their power is out.\nSo be careful: When the shields are down, you've only got your cannon left!\nTake into account that you won't earn points when you destroy stuff using your shields!\nEvery hit on your shields will decrease their power quite a bit!\n\n");
        helpForm.append("The TELEPORTER:\n");
        helpForm.append("It will move you to a random position and make you invincible for a short time. However, its use is limited. It takes a long time to regenerate!");
        helpForm.append("The (normal) POINTS:\n");
        helpForm.append("Small asteroid: 8\nMedium asteroid: 4\nBig asteroid: 2\nUFO: 20\nHoming missile: 40\n\n");
        helpForm.append("The LEVELS:\n");
        helpForm.append("This game has 10 levels. The only differences are the UFO and the number of asteroids appearing.");
        helpForm.append("The HIGHSCORES:\n");
        helpForm.append("If you should lose your last ship and your score is good enough for the high-score table, you can enter your initials after the GAME OVER screen.");
        helpForm.append("The CREATOR:\n");
        helpForm.append("Simon Lehmayr: I wrote this game because I missed the arcade game so much :-)\nSpecial thanks to Jean-Francois Doue for posting his mobile game engine (Ship, Asteroids, High-Scores and Title-Screen) under the GNU license.\nOf course I publish my source code for free (GNU license), too:\nhttp://desimon.funpic.de/Games/asteroids.zip\n\n");
        helpForm.append("TODO:\n- 2 Player version over BlueTooth\n\nPlease write me comments and suggestions at http://desimon.funpic.de/Games/suggest.php");

        _helpOkCommand = new Command("Ok", Command.SCREEN, 1);
        helpForm.addCommand(_helpOkCommand);
        helpForm.setCommandListener(this);
        
        // Instantiate the main game canvas.
        // Add two commands to the canvas, one to start the game
        // and one to exit the MIDlet
        field = new Field();
        _exitCommand = new Command("Exit", Command.EXIT, 1);
        _startCommand = new Command("Start", Command.SCREEN, 1);
        _helpCommand = new Command("Help", Command.SCREEN, 3);
        _muteCommand = new Command("Sound is off", Command.SCREEN, 4);
        _musicMuteCommand = new Command("Music is off", Command.SCREEN, 4);
        _difficultyCommand = new Command("Difficulty = normal", Command.SCREEN, 4);
        field.addCommand(_exitCommand);
        field.addCommand(_startCommand);
        field.addCommand(_muteCommand);
        field.addCommand(_musicMuteCommand);
        field.addCommand(_difficultyCommand);
        field.addCommand(_helpCommand);
        field.setCommandListener(this);
        _currentDisplayable = field;

        // Create a form which will be used to enter the
        // user name for high scores.
        scoreForm = new Form("Congratulations");
        _scoreField = new TextField("Your score is one of the 3 best. Enter your initials", "", 3, TextField.ANY);
        scoreForm.append(_scoreField);
        _scoreOkCommand = new Command("Done", Command.SCREEN, 1);
        scoreForm.addCommand(_scoreOkCommand);
        scoreForm.setCommandListener(this);

        // Read the high scores from persistent storage.
        scores = new Scores();

        // Initialize the random number generator.
        random = new Random();
    }

    /**
     * Method used to switch screens.
     */
    public static void setDisplayable(Displayable displayable) {
        _currentDisplayable = displayable;
        display.setCurrent(displayable);
    }

    /**
     * Invoked by Field to ask the game to switch to
     * the high score form.
     */
    public static void enterHighScore() {
        _scoreField.setString("");
        setDisplayable(scoreForm);
    }

    // Methods overwritten from MIDlet. These methods are
    // invoked by the application management software
    // to ask to MIDlet to enter the desired state.

    /**
     * Overriden from MIDlet.
     */
    protected void startApp() {
        // The game has two application threads
        // + One timer thread to trigger a switch between
        // the different game screens while not in game mode.
        // + A thread to run the main game loop.
        timer = new Timer();
        setDisplayable(_currentDisplayable);
//        System.out.println("Game :: startApp");
        Thread thread = new Thread(field);
        thread.start();
        field.setState(field.getState());
        sound = new Soundsystem();
        sound.start();
    }
    
    /**
     * Overriden from MIDlet.
     */
    protected void pauseApp() {
        // Kill the two game threads.
//        System.out.println("Game :: pauseApp");
        field.pause();
        if (timer != null) {
            timer.cancel();
            timer = null;
        }
    }
    
    /**
     * Overriden from MIDlet.
     */
    protected void destroyApp(boolean unconditional) throws MIDletStateChangeException {
//        System.out.println("Game :: destroyApp");
        sound.running = false;
        field.pause();
        timer.cancel();
        timer = null;
        random = null;
        field = null;
        scores = null;
        display = null; 
        _exitCommand = null;
        _startCommand = null;
        _scoreOkCommand = null;
        scoreForm = null;
        helpForm = null;
        _scoreField = null;
        _currentDisplayable = null;
        //System.out.println("computeavg = " + (Field.computeavg / Field.frames));
        //System.out.println("paintavg = " + (Field.paintavg / Field.frames));
    }
    
    // Implementation of the CommandListener interface
    /**
     * CommandListener interface implementation.
     */
    public void commandAction(Command c, Displayable d) {
        if (c == _exitCommand) {
            try {
                destroyApp(false);
            } catch(MIDletStateChangeException e) {}
            notifyDestroyed();
        } else if (c == _startCommand) {
            field.setState(Field.GAME_STATE);
            field.newGame();
        } else if (c == _scoreOkCommand) {
            scores.addHighScore(field.getScore(), _scoreField.getString());
            field.setState(Field.HISCORE_STATE);
            setDisplayable(field);
            field.setFrozen(false);
        } else if (c == _helpCommand) {
            field.setFrozen(true);
            setDisplayable(helpForm);
        } else if (c == _muteCommand) {
            sound.setMute(!sound.mute);
            field.removeCommand(_muteCommand);
            if (sound.mute) _muteCommand = new Command("Sound is off", Command.SCREEN, 4); else _muteCommand = new Command("Sound is on", Command.SCREEN, 4);
//            field.setItem(2, _muteCommand);
            field.addCommand(_muteCommand);
        } else if (c == _musicMuteCommand) {
            sound.setMusicMute(!sound.musicMute);
            field.removeCommand(_musicMuteCommand);
            if (sound.musicMute) _musicMuteCommand = new Command("Music is off", Command.SCREEN, 4); else _musicMuteCommand = new Command("Music is on", Command.SCREEN, 4);
//            field.setItem(2, _muteCommand);
            field.addCommand(_musicMuteCommand);
        } else if (c == _difficultyCommand) {
            field.setDifficulty((field.difficulty+1)%3);
            field.removeCommand(_difficultyCommand);
            _difficultyCommand = new Command("Difficulty = "+field.DIFFSTRING[field.difficulty], Command.SCREEN, 4);
//            field.setItem(3, _difficultyCommand);
            field.addCommand(_difficultyCommand);
        } else if (c == _helpOkCommand) {
            setDisplayable(field);
            field.setState(field.getState());
            field.setFrozen(false);
        }
    }
}

